/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (C) 2009  PaX Team <pageexec@freemail.hu> */

#ifndef _ASM_X86_ALTERNATIVE_ASM_H
#define _ASM_X86_ALTERNATIVE_ASM_H

#ifdef __ASSEMBLY__

#include <linux/linkage.h>
#include <asm/asm.h>
#include <asm/irq_vectors.h>
#include <asm/page.h>
#include <asm/percpu.h>

#ifdef CONFIG_SMP
	.macro LOCK_PREFIX
.Llock_prefix_\@:
	lock
	.pushsection .smp_locks,"a"
	.balign 4
	.long .Llock_prefix_\@ - .
	.popsection
	.endm
#else
	.macro LOCK_PREFIX
	.endm
#endif

/*
 * objtool annotation to ignore the alternatives and only consider the original
 * instruction(s).
 */
.macro ANNOTATE_IGNORE_ALTERNATIVE
	.Lannotate_\@:
	.pushsection .discard.ignore_alts
	.long .Lannotate_\@ - .
	.popsection
.endm

.macro pax_force_retaddr_bts rip=0
#ifdef KERNEXEC_PLUGIN
	btsq $63,\rip(%rsp)
#endif
.endm

#if defined(CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_BTS) && defined(CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_OR)
#error PAX: the KERNEXEC BTS and OR methods must not be enabled at once
#endif

#if defined(CONFIG_PAX_RAP_XOR) && defined(CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_OR)
#error PAX: RAP and the KERNEXEC OR methods must not be enabled at once
#endif

.macro pax_force_retaddr rip=0
#ifdef CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_BTS
	btsq $63,\rip(%rsp)
#endif
#ifdef CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_OR
	orq %r12,\rip(%rsp)
#endif
.endm

.macro pax_force_fptr ptr:req
#ifdef CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_BTS
	btsq $63,\ptr
#endif
#ifdef CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_OR
	orq %r12,\ptr
#endif
.endm

.macro pax_set_fptr_mask altstack=0
#ifdef CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_OR
	movabs $0x8000000000000000,%r12
#endif
#ifdef CONFIG_PAX_RAP_XOR
	.if \altstack
	movq PER_CPU_VAR(cpu_current_top_of_stack),%r12
	.else
	movq %rsp,%r12
	.endif
	andq $~(THREAD_SIZE_asm-1),%r12
	movq (%r12),%r12
#endif
.endm

.macro pax_jmp target:req
	jmp \target
#ifdef CONFIG_SLS
	int3
#endif
.endm

.macro pax_sls_call target:req
#ifdef CONFIG_SLS
	xor %eax, %eax
#endif
	call \target
.endm

#ifdef CONFIG_PAX_RAP_HASH
.macro misalign align:req offset:req pad=0xcc
#ifdef HAVE_MISALIGN
	.set .Loffset_\@,\offset
	.org .+(\align+(.Loffset_\@&(\align-1))-(.%\align))&(\align-1),\pad
#endif
.endm

.macro rap_retloc callee="" hashval=0 call_size=0
	.set .Lcall_size_\@,\call_size
	.set .Lrap_retloc_len_\@,.Lrap_retloc_end_\@-.Lrap_retloc_start_\@
//	.skip (16 - ((.Lrap_retloc_end_\@ - .Lrap_retloc_start_\@ - \call_size) % 16)) % 16, 0xcc
	misalign align=8 offset=-.Lrap_retloc_len_\@
.Lrap_retloc_start_\@:
	.ifnb \callee
	ASM_RAP_RET_HASH(__rap_hash_ret_\callee)
	.elseif \hashval
	ASM_RAP_RET_HASH(\hashval)
	.else
	.error "either 'callee' or 'hashval' must be specified"
	.endif
	.skip RAP_RET_HASH_OFFSET-RAP_HASH_LEN-.Lcall_size_\@,0xcc
.Lrap_retloc_end_\@:
.endm

.macro rap_call target:req hashsym="" hashval=0 sym="" lfence=0 sls=0 annotation=
//	.p2align 5,,2
	.if \lfence
	pax_jmp .Lrap_call_target_\@
	.else
	jmp .Lrap_call_target_\@
	.endif

	.ifnb \hashsym
	rap_retloc callee=\hashsym call_size=.Lrap_call_target_end_\@-.Lrap_call_target_\@
	.elseif \hashval
	rap_retloc hashval=\hashval call_size=.Lrap_call_target_end_\@-.Lrap_call_target_\@
	.else
	rap_retloc callee=\target call_size=.Lrap_call_target_end_\@-.Lrap_call_target_\@
	.endif

.Lrap_call_target_\@:
	.if \lfence
	lfence
	.endif

#ifdef CONFIG_SLS
	.if \sls
	xor %eax, %eax
	.endif
#endif

	.ifnb \sym
	.globl \sym
\sym :
	.endif

	\annotation
	call \target
.Lrap_call_target_end_\@:
.endm

.macro rap_ret func:req
	RET
.endm
#endif

.macro pax_indirect_jmp target:req
	pax_jmp *\target
.endm

.macro pax_direct_call_global target:req hashsym:req
#ifdef CONFIG_PAX_RAP_HASH
	rap_call \target, hashsym=\hashsym, sym=\hashsym sls=1
#else
#ifdef CONFIG_SLS
	xor %eax, %eax
#endif
	.globl \hashsym
\hashsym :
	call \target
#endif
.endm

.macro pax_indirect_call target:req hashsym:req lfence=0
#ifdef CONFIG_SLS
	.ifc \target,%_ASM_AX
	.else
	xor %eax, %eax
	.endif
#endif
#ifdef CONFIG_PAX_RAP_HASH
	rap_call *\target, hashsym=\hashsym, annotation=ANNOTATE_RETPOLINE_SAFE, lfence=\lfence
#else
	.if \lfence
	lfence
	.endif
	ANNOTATE_RETPOLINE_SAFE
	call *\target
#endif
.endm

.macro pax_direct_call target:req hashsym="" hashval=0 sls=0
#ifdef CONFIG_PAX_RAP_HASH
	rap_call \target, hashsym=\hashsym, hashval=\hashval, sls=\sls
#else
	.if \sls
	pax_sls_call \target
	.else
	call \target
	.endif
#endif
.endm

.macro pax_paravirt_call target:req hashsym:req
	pax_indirect_call \target, \hashsym
.endm

.macro pax_direct_sls_call target:req hashsym="" hashval=0
	pax_direct_call \target, hashsym=\hashsym, hashval=\hashval, sls=1
.endm

.macro pax_retloc callee:req sym=""
#ifdef CONFIG_PAX_RAP_HASH
	.ifnb \sym
	rap_hash.\sym:
	.endif
	rap_retloc callee=\callee
	.ifnb \sym
	.type rap_hash.\sym, @function
	.size rap_hash.\sym, . - rap_hash.\sym
	.endif
#else
#endif
.endm

.macro pax_ret func:req
	pax_force_retaddr
#ifdef CONFIG_PAX_RAP_HASH
	rap_ret \func
#else
	RET
#endif
.endm

.macro pax_ret_nospec
	RET
.endm

.macro pax_ret_nospec_alternative
#ifdef CONFIG_RETHUNK
	pax_jmp __x86_return_thunk
#else
	RET
#endif
.endm

/*
 * Issue one struct alt_instr descriptor entry (need to put it into
 * the section .altinstructions, see below). This entry contains
 * enough information for the alternatives patching code to patch an
 * instruction. See apply_alternatives().
 */
.macro altinstruction_entry orig alt feature orig_len alt_len pad_len prepad_len
	.long \orig - .
	.long \alt - .
	.word \feature
	.byte \orig_len
	.byte \alt_len
	.byte \pad_len
	.byte \prepad_len
.endm

/*
 * Define an alternative between two instructions. If @feature is
 * present, early code in apply_alternatives() replaces @oldinstr with
 * @newinstr. ".skip" directive takes care of proper instruction padding
 * in case @newinstr is longer than @oldinstr.
 */
.macro ALTERNATIVE oldinstr, newinstr, feature
# define old_len		.Lold_pad_\@-.Lold_\@
# define old_full_len		.Lold_end_\@-.Lold_\@
# define new_len		.Lnew_end_\@-.Lnew_\@
# define pad_len		.Lold_end_\@-.Lold_pad_\@
.Lold_\@\():
	\oldinstr
.Lold_pad_\@\():
	.skip -(((new_len)-(old_len)) > 0) * ((new_len)-(old_len)),0x90
.Lold_end_\@\():

	.pushsection .altinstructions,"a"
	altinstruction_entry .Lold_\@,.Lnew_\@,\feature,old_full_len,new_len,pad_len,0
	.popsection

	.pushsection .altinstr_replacement,"ax"
.Lnew_\@\():
	\newinstr
.Lnew_end_\@\():
	.popsection
# undef old_len
# undef old_full_len
# undef new_len
# undef pad_len
.endm

/*
 * gas compatible max based on the idea from:
 * http://graphics.stanford.edu/~seander/bithacks.html#IntegerMinOrMax
 *
 * The additional "-" is needed because gas uses a "true" value of -1.
 */
#define alt_max_short(a, b)	((a) ^ (((a) ^ (b)) & -(-((a) < (b)))))


/*
 * Same as ALTERNATIVE macro above but for two alternatives. If CPU
 * has @feature1, it replaces @oldinstr with @newinstr1. If CPU has
 * @feature2, it replaces @oldinstr with @feature2.
 */
.macro ALTERNATIVE_2 oldinstr, newinstr1, feature1, newinstr2, feature2
# define old_len		.Lold_pad_\@-.Lold_\@
# define old_full_len		.Lold_end_\@-.Lold_\@
# define new1_len		.Lnew1_end_\@-.Lnew1_\@
# define new2_len		.Lnew2_end_\@-.Lnew2_\@
# define pad1_len		.Lold_end_\@-.Lold_pad_\@
.Lold_\@\():
	\oldinstr
.Lold_pad_\@\():
	.skip -((alt_max_short(new1_len, new2_len) - (old_len)) > 0) * \
		(alt_max_short(new1_len, new2_len) - (old_len)),0x90
.Lold_end_\@\():

	.pushsection .altinstructions,"a"
	altinstruction_entry .Lold_\@,.Lnew1_\@,\feature1,old_full_len,new1_len,pad1_len,0
	altinstruction_entry .Lold_\@,.Lnew2_\@,\feature2,old_full_len,new2_len,pad1_len,0
	.popsection

	.pushsection .altinstr_replacement,"ax"
.Lnew1_\@\():
	\newinstr1
.Lnew1_end_\@\():

.Lnew2_\@\():
	\newinstr2
.Lnew2_end_\@:
	.popsection
# undef old_len
# undef old_full_len
# undef new1_len
# undef new2_len
# undef pad1_len
.endm

.macro ALTERNATIVE_2_PREPAD oldinstr, newinstr1, feature1, newinstr2, feature2
# define old_len		.Lold_end_\@-.Lold_\@
# define old_full_len		.Lold_end_\@-.Lold_pad_\@
# define new1_len		.Lnew1_end_\@-.Lnew1_\@
# define new1_full_len		.Lnew1_end_\@-.Lnew1_pad_\@
# define new2_len		.Lnew2_end_\@-.Lnew2_\@
# define new2_full_len		.Lnew2_end_\@-.Lnew2_pad_\@
# define prepad1_len		.Lnew1_\@-.Lnew1_pad_\@
# define prepad2_len		.Lnew2_\@-.Lnew2_pad_\@
.Lold_pad_\@\():
	.skip -((alt_max_short(new1_len, new2_len) - (old_len)) > 0) * \
		(alt_max_short(new1_len, new2_len) - (old_len)),0x90
.Lold_\@\():
	\oldinstr
.Lold_end_\@\():

	.pushsection .altinstructions,"a"
	altinstruction_entry .Lold_pad_\@,.Lnew1_pad_\@,\feature1,old_full_len,new1_full_len,0,prepad1_len
	altinstruction_entry .Lold_pad_\@,.Lnew2_pad_\@,\feature2,old_full_len,new2_full_len,0,prepad2_len
	.popsection

	.pushsection .altinstr_replacement,"ax"
.Lnew1_pad_\@\():
	.skip -((alt_max_short(new2_len, old_len) - (new1_len)) > 0) * \
		(alt_max_short(new2_len, old_len) - (new1_len)),0x90
.Lnew1_\@\():
	\newinstr1
.Lnew1_end_\@\():

.Lnew2_pad_\@\():
	.skip -((alt_max_short(new1_len, old_len) - (new2_len)) > 0) * \
		(alt_max_short(new1_len, old_len) - (new2_len)),0x90
.Lnew2_\@\():
	\newinstr2
.Lnew2_end_\@\():
	.popsection
# undef old_len
# undef old_full_len
# undef new1_len
# undef new1_full_len
# undef new2_len
# undef new2_full_len
# undef prepad1_len
# undef prepad2_len
.endm

.macro __PAX_REFCOUNT section, counter
#ifdef CONFIG_PAX_REFCOUNT
	jo .Lrefcount_error_\@
	.pushsection .text.\section\().o
.Lrefcount_error_\@:
	lea \counter,%_ASM_CX
.Lrefcount_report_\@:
	int $X86_REFCOUNT_VECTOR
	.popsection
.Lrefcount_continue_\@:
	_ASM_EXTABLE(.Lrefcount_report_\@, .Lrefcount_continue_\@)
#endif
.endm

.macro PAX_REFCOUNT64_OVERFLOW counter
	__PAX_REFCOUNT refcount64_overflow, \counter
.endm

.macro PAX_REFCOUNT64_UNDERFLOW counter
	__PAX_REFCOUNT refcount64_underflow, \counter
.endm

/* If @feature is set, patch in @newinstr_yes, otherwise @newinstr_no. */
#define ALTERNATIVE_TERNARY(oldinstr, feature, newinstr_yes, newinstr_no)	\
	ALTERNATIVE_2 oldinstr, newinstr_no, X86_FEATURE_ALWAYS,		\
	newinstr_yes, feature

#endif  /*  __ASSEMBLY__  */

#endif /* _ASM_X86_ALTERNATIVE_ASM_H */
